#! python3
# KNX communication
# (C) 2017-2018 Dietmar Schwertberger


import collections
import asyncio
import time, struct

from KNXUtils import *

class KNX:
    def __init__(self, address=None, try_local=False, auto_reconnect=True, monitor_eibd=True, log_messages=None):
        self.address = address
        self.try_local = try_local

        #self.lock = threading.Lock()
        self.cache = {}
        self.datatypes = {} # GA:dpt
        self.names = {} # GA:name
        self.GAs = {}   # name:GA
        self.read_cached = set()
        self.queue = collections.deque()
        self.handlers = {} # GA:callback
        self._raw_handlers = []

        # for error logging
        self._last_sent     = None # (time.time(), GA, name, value)
        self._last_received = None # (time.time(), GA, name, value)
        self._last_eibd_check = None
        
        self._auto_reconnect = auto_reconnect
        self._monitor_eibd = monitor_eibd
        self._log_messages = log_messages

        self._thread_ident = None

        # asyncio
        self.writer = None
        self._connect_request_event = asyncio.Event()
        self._lock = asyncio.Lock()
        
        self.state = None # None, "connected", "error", "closed"
        self.connected = asyncio.Event()
        self.connected.clear()
        #self.received = asyncio.Event()
        #self.received.clear() # set on each telegram reception
        ##################################################################################
        # XXX
        # to be done: use coroutine for the sender synchronization
        #             this should try to restrict the number of telegrams waiting for reply
        #             during connection start
        self.message_acknowledged = asyncio.Event()
        # self.message_acknowledged.clear()
        # yield from self.message_acknowledged.wait()
        # self.message_acknowledged.set() # continue
        # self.message_acknowledged.notify()
        # self.message_acknowledged.release()
    def __bool__(self):
        #return self.state=="connected"
        return self.connected.is_set()
    def check_eibd(self):
        # check whether eibd is running; re-start if not
        if not self._monitor_eibd: return
        if self._last_eibd_check is not None and time.time()-self._last_eibd_check<10.0:
            return
        self._last_eibd_check = time.time()
        import subprocess
        ps = subprocess.Popen(('ps', 'cax'), stdout=subprocess.PIPE)
        output = ps.communicate()[0]
        splitted = output.split(b'\n')
        COMMAND = splitted[0].index(b"COMMAND") # F   UID   PID  PPID PRI  NI    VSZ   RSS WCHAN  STAT TTY        TIME COMMAND
        for line in splitted[1:]:
            if line[COMMAND:].startswith(b"eibd") and not b"<defunct>" in line:
                print("eibd running")
                return None
        # eibd not in process list: re-start
        try:
            eibd = subprocess.Popen(["eibd","-i","-D","-T","-S","-e 1.1.251", "tpuarts:/dev/ttyAMA0"])
            if self._log_messages:
                #self._log_messages("restarted eibd as pid %d\n%s\n\n"%(eibd.pid, output.decode("latin1")) )
                self._log_messages("restarted eibd as pid %d\n"%eibd.pid)
            return True
        except FileNotFoundError:
            if self._log_messages:
                self._log_messages("could not re-start eibd: FileNotFoundError")
            return False
    def connect(self, address=None):
        if address is not None and address!=self.address:
            self._close()
            self.address = address
        if self._auto_reconnect:
            return
        self._connect_request_event.set()
    def on_connect(self):
        pass
    def on_disconnect(self):
        pass
    def close(self):
        self._close()
    @asyncio.coroutine
    def mainloop(self):
        
        # XXX change to asyncio queue for sending
        # tasks = asyncio.shield(receiving), self.queue.
        #  or better:
        # tasks = receiving, self.queue
        # done, pending = asyncio.wait(tasks, timeout=2.0, return_when=asyncio.FIRST_COMPLETED) # FIRST_EXCEPTION ALL_COMPLETED

        # in addition to mainloop:
        #  one coroutine and queue for receiving, one queue for sending
        print("Entering KNX mainloop")
        import threading
        self._thread_ident = threading.current_thread().ident
        while True:
            if not self._auto_reconnect:
                yield from self._connect_request_event.wait() # wait until self.connect() is called
                self._connect_request_event.clear()
            print( "KNX connecting")
            reader = None
            if self.try_local:
                try:
                    reader, writer = yield from asyncio.wait_for( asyncio.open_connection("127.0.0.1", 6720), 60.0 )
                except:
                    print("KNX could not connect to local eibd")
            try: # connect with timeout
                reader, writer = yield from asyncio.wait_for( asyncio.open_connection(self.address, 6720), 60.0 )
                #self.state = "connected"
                print( "KNX connected" )
                self.try_local = False # don't try any more on connection lossself.on_connect()
            except asyncio.TimeoutError:
                print("KNX connect timeout")
                if self._monitor_eibd:
                    result = self.check_eibd()
                    if result is False: yield from asyncio.sleep(1.0)
                if not self._auto_reconnect:
                    self._close()
                    return
                continue
            except:
                print("KNX connect exception")
                if self._monitor_eibd:
                    result = self.check_eibd()
                    if result is False: yield from asyncio.sleep(1.0)
                    if not self._auto_reconnect:
                        self._close()
                        return
                continue
            
            assert isinstance(reader, asyncio.streams.StreamReader)

            # connected; read and write; in case of error etc., close the connection and 
            self.cache.clear()
            self.state = "connected"
            self.connected.set()
            self.writer = writer
            self.queue.clear()
            self.queue.append( (build_telegram("H", [EIB_CACHE_ENABLE]), None) )
            for GA in self.read_cached:
                self.queue.append( (build_telegram( "HHBB", [EIB_CACHE_READ,decode_GA_string(GA),0,KNXREAD] ), None) )
            self.queue.append( (build_telegram( "HHB", [EIB_OPEN_GROUPCON,0,0] ),None) )
            self.trigger_send()

            self.on_connect()
            
            receiving = None # the task to receive two bytes
            while True:
                # communicate main loop
                try:
                    # try to read header
                    if receiving is None:
                        receiving = asyncio.Task( reader.readexactly(2) )
                    # shield the task not to lose bytes
                    rx = yield from asyncio.wait_for( asyncio.shield(receiving), 2.0)
                    receiving = None
                    if not rx:
                        # remote has closed connection
                        self._close("closed")
                        break
                except asyncio.TimeoutError:
                    rx = None
                except:
                    print("KNX execption")
                    break

                if rx is not None:
                    # read rest of telegram
                    length = struct.unpack(">H", rx)[0]
                    try:
                        telegram = yield from asyncio.wait_for( reader.readexactly(length), 0.5)
                    except:
                        # log error together with info about last sent and received telegrams
                        now = time.time()
                        tx = self._last_sent
                        if tx: # replace timestamp with time delta
                            tx = (now-tx[0],) + tx[1:]
                        rx = self._last_received
                        if rx: # replace timestamp with time delta
                            rx = (now-rx[0],) + rx[1:]
                        
                        if self._log_messages:
                            self._log_messages( "KNX telegram receive failed (%d bytes);\n   last_sent: %s\n   last_received: %s"%(length,repr(tx),repr(rx) ) )
                        if WINDOWS:
                            raise
                        break
                    #self.received.set()
                    try:
                        self.handle_telegram(telegram)
                    except:
                        raise

                if self.queue:
                    self.trigger_send()

            if self.writer is not None:
                # the main loop was left with just a break; this only happens on a fatal exception
                self._close("error")
            
    def _close(self, state="closed"):
        print("XXXXXXXXXXXXXXXXXXXX CLOSING KNX")
        #self.state = state
        self.connected.clear()
        self.state = state
        writer = self.writer
        if writer is None: return
        self.writer = None
        writer.close()
        self.queue.clear()
        self.on_disconnect()
        if self._log_messages:
            self._log_messages("KNX connection closed: %s"%state)
    def reconnect(self):
        pass
    def trigger_send(self):
        if not self.queue or not self.writer: return
        d, logging_info = self.queue.popleft()
        print("KNX >", d)
        self.writer.write(d)
        self._last_sent = logging_info

    def handle_telegram(self, telegram):
        # on receipt

        # 2 bytes type
        # 2 bytes src
        # 2 bytes dst
        # 2 bytes command/data
        # rest: data
        ttype = struct.unpack_from(">H", telegram)[0]
        if ttype!=EIB_GROUP_PACKET and ttype!=EIB_CACHE_READ:
            # unknown type
            return
        if len(telegram)>=8:
            source, destination,DRL,flag = struct.unpack_from(">HHBB", telegram, 2)
        elif len(telegram)==6:
            source, destination = struct.unpack_from(">HH", telegram, 2)
            DRL = 0
            flag = 0
        else:
            raise ValueError( "XXX" )
        source = encode_PA_string(source) # always a physical address
        # DRL: destination-address-flag, routing-counter, length: DRRRLLLL
        if DRL: # 17th bit of destination set: it's a physical address
            destination = encode_PA_string(destination)
        else: # group address
            destination = encode_GA_string(destination)
        flag = FLAGS[flag&0xC0]
        data = telegram[7:]
        
        knx_dpt = decoded_data = None
    
        # find data type (dpt) for some GAs
        knx_dpt = self.datatypes.get(destination)
        if knx_dpt is None and len(data)==1:
            knx_dpt = 1

        # decode
        if knx_dpt is not None:
            decoded_data = decode(data, knx_dpt)
            if decoded_data is None:
                print( "*********** None", destination,self.names.get(destination), repr(data) )
            self.cache[destination] = decoded_data

        name = self.names.get(destination)
        name_ = "KNX: %s:"%name if name else "KNX:"
        if not "Heartbeat" in name_:
            if decoded_data is not None:
                # could be decoded
                print( "%-35s"%name_, hex(ttype), flag, "%-12s"%source, "%-14s"%destination, "decoded", decoded_data )
            else:
                print( "%-35s"%name_, hex(ttype), flag, "%-12s"%source, "%-14s"%destination, repr(telegram[7:]), "%d bytes"%len(telegram[7:]) )

        self._last_received = (time.time(), destination, name, decoded_data)

        # call handlers
        for callback in self._raw_handlers + self.handlers.get(destination, []):
            #ret = callback(destination, name, decoded_data)
            try:
                ret = callback(destination, decoded_data)
            except:
                if self._log_messages:
                    import traceback, sys
                    exception = sys.exc_info()[0]
                    exception = getattr( exception, "__name__", str(exception) )
                    message  = "Exception in handler, called from KNX.handle_telegram:\n%s: %s\n\nTraceback information:\n"%(exception, sys.exc_info()[1])
                    message += "\n".join(traceback.format_tb(sys.exc_info()[2]))
                    self._log_messages(message)
                    print(message)

    ######################################################################################
    def register(self, GA, dpt, name=None, read_cached=False, callback=None):
        if name is not None:
            self.names[GA] = name
            self.GAs[name] = GA
        self.datatypes[GA] = dpt
        if callback is not None:
            self.register_handler(GA, callback)
        if read_cached:
            self.read_cached.add(GA)
    def register_handler(self, GA, callback):
        # callback will be called back with (GA, name, value)
        GA = self.GAs.get(GA,GA)
        self.handlers.setdefault(GA, []).append(callback)
    def register_raw_handler(self, callback):
        # for raw data, e.g. to be forwarded via network
        self._raw_handlers.append( callback )
    def send(self, GA, value, dpt=None):
        if not self: return False
        GA = self.GAs.get(GA,GA)
        if dpt is None:
            dpt = self.datatypes[GA]
        GA = decode_GA_string(GA)
        flag, data = encode(value, dpt)
        
        #if dpt in (1,3):
        telegram = build_telegram( "HHBB", [EIB_GROUP_PACKET,GA,0, flag], data ) # data included
        logging_info = (time.time(), GA, self.GAs.get(GA), value)
        trigger_send = len(self.queue)<3
        self.queue.append( (telegram, logging_info) )
        if trigger_send:
            # XXX change to asyncio queue
            import threading
            if threading.current_thread().ident == self._thread_ident:
                self.trigger_send()
        return True
    def get(self, GA, default=None):
        # read cached value; return default if not cached
        GA = self.GAs.get(GA,GA)
        return self.cache.get(GA, default)

