#! python3

try:
    basestring
except:
    basestring = str

EIB_INVALID_REQUEST            =0x00
EIB_CONNECTION_INUSE           =0x01
EIB_PROCESSING_ERROR           =0x02
EIB_CLOSED                     =0x03
EIB_RESET_CONNECTION           =0x04

EIB_OPEN_BUSMONITOR            =0x10
EIB_OPEN_BUSMONITOR_TEXT       =0x11
EIB_OPEN_VBUSMONITOR           =0x12
EIB_OPEN_VBUSMONITOR_TEXT      =0x13
EIB_BUSMONITOR_PACKET          =0x14

EIB_OPEN_T_CONNECTION          =0x20
EIB_OPEN_T_INDIVIDUAL          =0x21
EIB_OPEN_T_GROUP               =0x22
EIB_OPEN_T_BROADCAST           =0x23
EIB_OPEN_T_TPDU                =0x24
EIB_APDU_PACKET                =0x25
EIB_OPEN_GROUPCON              =0x26
EIB_GROUP_PACKET               =0x27 # 39

EIB_PROG_MODE                  =0x30
EIB_MASK_VERSION               =0x31
EIB_M_INDIVIDUAL_ADDRESS_READ  =0x32

EIB_M_INDIVIDUAL_ADDRESS_WRITE =0x40
EIB_ERROR_ADDR_EXISTS          =0x41
EIB_ERROR_MORE_DEVICE          =0x42
EIB_ERROR_TIMEOUT              =0x43
EIB_ERROR_VERIFY               =0x44

EIB_MC_INDIVIDUAL              =0x49
EIB_MC_CONNECTION              =0x50
EIB_MC_READ                    =0x51
EIB_MC_WRITE                   =0x52
EIB_MC_PROP_READ               =0x53
EIB_MC_PROP_WRITE              =0x54
EIB_MC_PEI_TYPE                =0x55
EIB_MC_ADC_READ                =0x56
EIB_MC_AUTHORIZE               =0x57
EIB_MC_KEY_WRITE               =0x58
EIB_MC_MASK_VERSION            =0x59
EIB_MC_RESTART                 =0x5a
EIB_MC_WRITE_NOVERIFY          =0x5b
EIB_MC_PROG_MODE               =0x60
EIB_MC_PROP_DESC               =0x61
EIB_MC_PROP_SCAN               =0x62
EIB_LOAD_IMAGE                 =0x63

EIB_CACHE_ENABLE               =0x70
EIB_CACHE_DISABLE              =0x71
EIB_CACHE_CLEAR                =0x72
EIB_CACHE_REMOVE               =0x73
EIB_CACHE_READ                 =0x74
EIB_CACHE_READ_NOWAIT          =0x75
EIB_CACHE_LAST_UPDATES         =0x76


KNXREAD = 0x00
KNXRESP = 0x40
KNXWRITE = 0x80
FLAGS = {KNXREAD:"read", KNXWRITE:"write", KNXRESP:"response"}

"""Data types:           
EIS: European Internetworking Standard (old)
DPT: KNX Data Point Type (new)
EIS 1 Schalten                                1 Bit     DPT 1
EIS 2 Dimmen (Position, Control, Wert)        1+4+8 bit DPT 3
EIS 3 Uhrzeit                                 3 Byte    DPT 10
EIS 4 Datum                                   3 Byte    DPT 11
EIS 5 Gleitkommawert                          2 Byte    DPT9
EIS 6 Relativwert, 0-255%                     1 Byte    BPT5, 6
EIS 7 Antriebssteuerung Jalousie/Rollladen    1 Bit     DPT1
EIS 8 Zwangssteuerung                         1+2 Bit   DPT2
EIS 9 Gleitkommawert IEEE                     4 Byte    DPT14
EIS 10 16-Bit Wert                            2 Byte    DPT7,8
EIS 11 32-Bit Wert                            4 Byte    DPT 12,13
EIS 12 Zugangskontrolle                       1 Byte    DPT 15
EIS 13 ASCII Zeichen                          1 Byte    DPT4
EIS 14 8-Bit Wert                             1 Byte    DPT5,6
EIS 15 Zeichenkette

by DPT:
EIS: European Internetworking Standard (old)
DPT: KNX Data Point Type (new)
EIS 1 Schalten                                1 Bit     DPT 1
EIS 7 Antriebssteuerung Jalousie/Rollladen    1 Bit     DPT1
EIS 8 Zwangssteuerung                         1+2 Bit   DPT2
EIS 2 Dimmen (Position, Control, Wert)        1+4+8 bit DPT 3
EIS 13 ASCII Zeichen                          1 Byte    DPT4
EIS 14 8-Bit Wert                             1 Byte    DPT5,6
EIS 6 Relativwert, 0-255%                     1 Byte    BPT5,6
EIS 10 16-Bit Wert                            2 Byte    DPT7,8
EIS 5 Gleitkommawert                          2 Byte    DPT9
EIS 3 Uhrzeit                                 3 Byte    DPT 10
EIS 4 Datum                                   3 Byte    DPT 11
EIS 11 32-Bit Wert                            4 Byte    DPT 12,13
EIS 9 Gleitkommawert IEEE                     4 Byte    DPT14
EIS 12 Zugangskontrolle                       1 Byte    DPT 15
EIS 15 Zeichenkette


"""

import struct

def decode_GA_string(GA):
    # group address string "X/Y/Y" to integer (16 bit)
    if isinstance(GA, int): return GA
    #GA = "2/6/11" # 16 bits, 15 used; written as 4/3/8 bits
    splitted = GA.split("/")
    if len(splitted)==3:
        main,med,sub = splitted
        return (int(main)<<11) + (int(med)<<8) + int(sub)
    elif len(splitted)==2:
        main,sub = splitted
        return (int(main)<<11) + int(sub)
    raise ValueError( "wrong GA string format; either 'M/N/O' or 'M/O' required" )

def encode_GA_string(GA, length=3):
    # group address string "X/Y/Y" to integer (16 bit)
    if isinstance(GA, basestring): return GA
    #GA = "2/6/11" # 16 bits, 15 used; written as 4/3/8 bits
    sub = 0xFF & GA
    if length==3:
        #med  = (0b11100000000&GA) >> 8
        med  = (0x700&GA) >> 8
        #main = (0b111100000000000&GA) >> 11
        main = (0x7800&GA) >> 11
        return "%d/%d/%d"%(main,med,sub)
    raise ValueError( "not yet implemented" )

def encode_PA_string(PA):
    # group address string "X/Y/Y" to integer (16 bit)
    if isinstance(PA, basestring): return PA
    #GA = "2/6/11" # 16 bits, 15 used; written as 4/3/8 bits
    teilnehmer = 0xFF & PA
    #line  =    (0b111100000000&PA) >> 8
    line  =    (0xF00&PA) >> 8
    #main = (0b1111000000000000&PA) >> 12
    main = (0xF000&PA) >> 12
    return "%d.%d.%d"%(main,line,teilnehmer)

def encode_EIBD_data(data, dpt=1):
    if dpt==1:
        return "B", 0x80 | data
    
    pass

def build_telegram(fmt, data):
    # big endian
    ret = struct.pack(">"+fmt, *data)
    if len(ret) < 2 or len(ret) > 0xffff:
        raise ValueError("(encoded) data length needs to be between 2 and 65536")
    # prepend data length
    return struct.pack(">H", len(ret)) + ret

def build_telegram(fmt, header, data=""):
    # big endian
    ret = struct.pack(">"+fmt, *header)
    if data:
        ret += data
    if len(ret) < 2 or len(ret) > 0xffff:
        raise ValueError("(encoded) data length needs to be between 2 and 65536")
    # prepend data length
    return struct.pack(">H", len(ret)) + ret

def decode_dpt1(data):
    if len(data) != 1:
        return None
    if not data: return None
    return ord(data) & 0x3F # 0x80/0x81 -> 0/1 for switching

def encode_dpt1(value):
    value = 1 if value else 0
    #return struct.pack(">B", KNXWRITE | value)
    return KNXWRITE | value

def encode_dpt3(direction, step):
    direction = (direction & 0b01) <<3
    step      = (step & 0b0111)
    #return struct.pack(">B", KNXWRITE | direction | step)
    return KNXWRITE | direction | step

def decode_dpt3(data):
    if len(data) != 1:
        return None
    data = ord(data)
    up_down = data>>3 & 0b01
    step    = data & 0b0111
    return up_down, step

def decode_dpt5(data):
    # a (unsigned) byte
    if not data: return None
    return struct.unpack('>B', data)[0]

def encode_dpt5(value):
    # a (unsigned) byte
    return struct.pack('>B', value)

def encode_dpt7(value):
    return struct.pack('>H', value)

def decode_dpt9(data):
    if len(data) != 2:
        return None
    i1, i2 = struct.unpack(">BB", data)
    #i1 = ord(data[0])
    #i2 = ord(data[1])
    s = (i1 & 0x80) >> 7
    e = (i1 & 0x78) >> 3
    m = (i1 & 0x07) << 8 | i2
    if s == 1:
        s = -1 << 11
    return (m | s) * 0.01 * pow(2, e)

def encode_dpt9(value):
    # floating point
    s = 0
    e = 0
    if value < 0:
        s = 0x8000
    m = int(value * 100)
    while (m > 2047) or (m < -2048):
        e = e + 1
        m = m >> 1
    num = s | (e << 11) | (int(m) & 0x07ff)
    return struct.pack('>H', num)

def decode_dpt10(data):
    if len(data) != 3:
        return None
    hour, minute, second = struct.unpack('BBB', data)
    return (hour & 0x1F, minute & 0x3F, second & 0x3F)

def decode_dpt11(data):
    if len(data) != 3:
        return None
    day, month, year = struct.unpack('BBB', data)
    return (2000 + year & 0x7F, month & 0x0F, day & 0x1F) # 0b1111111, 0b1111, 0b11111 # 7,4,5 bits

def encode_dpt10(data=None):
    # (hour,minute,second) or (weekday [1-7],hour,minute,second) or time.localtime() tuple
    if data is None: data = time.localtime()
    weekday = 0 # no day, otherwise 1 to 7 for monday to sunday
    if len(data)== 3:
        hour,minute,second = data
    elif len(data)==4:
        weekday,hour,minute,second = data
    else:
        hour,minute,second = data[3:6]
        weekday = data.tm_wday + 1
    return struct.pack( ">BBB", weekday<<5 | hour & 0x1F, minute & 0x3F, second & 0x3F)


def encode_dpt11(data=None):
    # (year, month, day) or time.localtime() tuple
    if data is None: data = time.localtime()
    year, month, day = data[:3]
    if year>=2000:
        year = year - 2000
    elif year>=1990:
        year = year - 1990
    elif year > 99:
        raise ValueError("invalid year")
    return struct.pack( '>BBB', day & 0x1F, month & 0x0F, year & 0x7F) # 5, 4, 7 bits


def decode_GLT(data):
    # Griesser shutter object; angle is not used as we have only shutters
    if len(data)==4:
        position, angle, states, other_states = struct.unpack("BBBB", data)
        decoded_data = (position, states, other_states)
        return decoded_data
    if not data:
        return None
    return "ERROR - wrong length"

def decode(data, dpt=1):
    if dpt==1:
        return decode_dpt1(data)
    if dpt==3:
        return decode_dpt3(data)
    data = data[1:]
    if dpt==5:
        return decode_dpt5(data)
    if dpt==9:
        return decode_dpt9(data)
    if dpt==10:
        return decode_dpt10(data)
    if dpt==11:
        return decode_dpt11(data)
    if dpt=="GLT":
        return decode_GLT(data)
    raise ValueError( "unknown data type" )


def encode(data, dpt=1):
    if dpt==1:
        return encode_dpt1(data), ""
    if dpt==3:
        return encode_dpt3(data), ""
    if dpt==5:
        return KNXWRITE, encode_dpt5(data)
    if dpt==9:
        return KNXWRITE, encode_dpt9(data)
    if dpt==10:
        return KNXWRITE, encode_dpt10(data)
    if dpt==11:
        return KNXWRITE, encode_dpt11(data)
    raise ValueError( "unknown data type" )

